from rest_framework.authentication import BaseAuthentication, get_authorization_header
from rest_framework.exceptions import AuthenticationFailed

from .middleware import get_user

INVALID_TOKEN_MESSAGE = "Invalid JWT Token"


class ShopifyTokenAuthentication(BaseAuthentication):
    keyword = "Bearer"

    def authenticate(self, request):
        auth = get_authorization_header(request).split()
        if not auth or auth[0].lower() != self.keyword.lower().encode():
            return None
        if len(auth) == 1:
            msg = "Invalid token header. No credentials provided."
            raise AuthenticationFailed(msg)
        elif len(auth) > 2:
            msg = "Invalid token header. Token string should not contain spaces."
            raise AuthenticationFailed(msg)

        try:
            token = auth[1].decode()
        except UnicodeError:
            msg = "Invalid token header. Token string should not contain invalid characters."
            raise AuthenticationFailed(msg)

        user = get_user(token)
        if not user:
            raise AuthenticationFailed(INVALID_TOKEN_MESSAGE)

        return user, token



import logging
import re
from urllib.parse import urlparse
from django.core.exceptions import ObjectDoesNotExist
from django.contrib.auth import get_user_model

from django.conf import settings
from jose import jwt
from jose.exceptions import ExpiredSignatureError, JWTClaimsError, JWTError


def get_hostname(url):
    return urlparse(url).netloc

def decode_token(token):
    decoded_payload = jwt.decode(
        token,
        settings.SHOPIFY_APP_API_SECRET,
        algorithms=["HS256"],
        audience=settings.SHOPIFY_APP_API_KEY,
        options={"verify_sub": False, "verify_nbf": False},
    )
    dest_host = get_hostname(decoded_payload["dest"])
    iss_host = get_hostname(decoded_payload["iss"])

    assert (
        dest_host == iss_host
    ), "'dest' claim host does not match 'iss' claim host"
    assert dest_host, "'dest' claim host not a valid shopify host"

    return dest_host

def get_user(token):
    try:
        dest_host = decode_token(token)
        return get_user_model().objects.get(myshopify_domain=dest_host)

    except (
        ExpiredSignatureError,
        JWTError,
        JWTClaimsError,
        AssertionError,
        ObjectDoesNotExist,
        KeyError,
    ) as e:
        logging.warning(f"Login user failed: {e}.")

def authenticate(request):
    jwt_token = request.headers.get("Authorization")

    if not jwt_token:
        return

    striped_jwt_token = re.sub(r"Bearer\s", "", jwt_token)
    return get_user(striped_jwt_token)


class SessionTokensAuthMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        user = authenticate(request)
        if user:
            request.user = user
        return self.get_response(request)



import logging
import base64
import urllib.parse
from django.conf import settings
from django.http.response import HttpResponse, HttpResponseRedirect
from django.shortcuts import render, reverse
from django.views.generic.base import View

import shopify

from django.contrib.auth import get_user_model


def base64_decode(string):
    """
    Adds back in the required padding before decoding.
    """
    padding = 4 - (len(string) % 4)
    string = string + ("=" * padding)
    return base64.urlsafe_b64decode(string)


def get_scope_permission(request, myshopify_domain):
    redirect_uri = request.build_absolute_uri(reverse("session_tokens:finalize"))
    myshopify_domain = myshopify_domain.strip()
    permission_url = shopify.Session(
        myshopify_domain,
        getattr(settings, "SHOPIFY_APP_API_VERSION", "unstable"),
    ).create_permission_url(settings.SHOPIFY_APP_API_SCOPE, redirect_uri)

    embedded = request.GET.get("embedded")
    if not embedded or embedded != "1":
        return HttpResponseRedirect(permission_url)

    host = request.GET.get("host")
    if not host:
        raise Exception("expected host query parameter when embedded=1 is present")

    return render(
        request,
        "shopify_auth/iframe_redirect.html",
        {"host": host, "redirect_uri": permission_url},
    )


class FinalizeAuthView(View):
    def get(self, request):
        myshopify_domain = request.GET.get("shop")

        try:
            shopify_session = shopify.Session(
                myshopify_domain,
                getattr(settings, "SHOPIFY_APP_API_VERSION", "unstable"),
            )
            shopify_session.request_token(request.GET)
        except:
            logging.exception("Shopify login failed.")
            return HttpResponse("Shopify login failed.")

        shop = get_user_model().update_or_create(shopify_session, request)
        shop.install(request)

        if host := request.GET.get("host"):
            decoded_host = base64_decode(host).decode('utf-8')
            redirect_url = f"https://{decoded_host}/apps/{settings.SHOPIFY_APP_API_KEY}"
            return HttpResponseRedirect(redirect_url)


        return HttpResponseRedirect(
            f"https://{myshopify_domain}/admin/apps/{settings.SHOPIFY_APP_API_KEY}"
        )



from django.test import TestCase, Client
from django.conf import settings


class ViewsTestCase(TestCase):

    def setUp(self):
        self.client = Client()

    def tearDown(self):
        settings.SHOPIFY_APP_DEV_MODE = False
        self.client = None

    def test_login_view(self):
        """
        Test the login view loads when we're an anonymous user.
        """
        response = self.client.get('/')
        self.assertEqual(response.status_code, 200)

    def test_authenticate_view(self):
        """
        Test the authenticate view renders correctly with a shop param.
        """
        response = self.client.get('/authenticate/')
        self.assertEqual(response.status_code, 302)
        response = self.client.get('/authenticate/', follow=True)
        self.assertEqual(response.status_code, 404)
        response = self.client.post('/authenticate/')
        self.assertEqual(response.status_code, 302)

        # Dev mode so token does not need to be valid
        settings.SHOPIFY_APP_DEV_MODE = True
        response = self.client.get('/authenticate/?shop=test.myshopify.com')
        self.assertEqual(response.status_code, 302)
        self.assertGreater(int(self.client.session['_auth_user_id']), 0)
        self.assertEqual(self.client.session['_auth_user_backend'], 'shopify_auth.backends.ShopUserBackend')
        self.assertIsNot(self.client.session['_auth_user_hash'], None)

    def test_authenticates_standalone_app_with_shop_param(self):
        response = self.client.get('/?shop=test.myshopify.com')
        self.assertEqual(response.status_code, 302)

    def test_redirect_to_view(self):
        """
        Test that return_address is persisted through login flow.
        """
        response = self.client.get('/authenticate/?shop=test.myshopify.com&next=other-view')
        self.assertRedirects(response, 'https://test.myshopify.com/admin/oauth/authorize?client_id=test-api-key&scope=read_products&redirect_uri=http%3A%2F%2Ftestserver%2Ffinalize%2F', fetch_redirect_response=False)

        settings.SHOPIFY_APP_DEV_MODE = True
        response = self.client.get('/authenticate/?shop=test.myshopify.com')
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, 'other-view')
